#!/usr/bin/perl
# Author: Alex Efros <powerman-asdf@yandex.ru>, 2007,2008
# License: Public Domain
#
# asciidoc.cgi: View your AsciiDoc-formatted documentation in browser.
#
# This script will automatically convert your *.txt files into *.html
# on-the-fly when user send request to your website, so you can just
# update *.txt file and immediately view changes on website.
#
# You need to set up 3 directories on your website:
# 
# - Directory which should be used to view documentation on website.
#   Create empty directory for this and put there:
#   1) asciidoc.cgi
#   2) .htaccess:
#       ErrorDocument 404 /doc/doc.cgi
#       DirectoryIndex /doc/doc.cgi
#   3) You may need default images and stylesheets from AsciiDoc distribution:
#       cp -r /etc/asciidoc/images/                     /THIS/DIRECTORY/
#       cp /etc/asciidoc/stylesheets/xhtml11.css        /THIS/DIRECTORY/
#       cp /etc/asciidoc/stylesheets/xhtml11-quirks.css /THIS/DIRECTORY/
#       cp /etc/asciidoc/javascripts/toc.js             /THIS/DIRECTORY/
#   Default path (relative to website document root): /doc/
# 
# - Directory with your AsciiDoc-formatted documentation files.
#   All *.txt files in this directory and it subdirectories become
#   available for viewing on your website with help of asciidoc.cgi.
#   Default path (relative to asciidoc.cgi): ../.lib/doc/
# 
# - Directory where AsciiDoc will cache your *.txt converted to *.html.
#   This directory should be writable by asciidoc.cgi, so you may need to
#   make it world-writable.
#   Default path (relative to asciidoc.cgi): ../.lib/tmp/doc/
# 
# Probably you'll wish to disable stright access to last two directories
# from website, so put there .htaccess like this:
#       order allow,deny
#       allow from none
#       deny from all
# 
# If you'll wanna change default paths for these directories - just edit
# "Configuration" section below in code.
# 
# Now you can create .lib/doc/index.txt and view it using any of these URLs:
#       http://your.website/doc/
#       http://your.website/doc/index
#       http://your.website/doc/index.htm
#       http://your.website/doc/index.html

use warnings;
use strict;
use File::Spec;
use Fcntl qw(:DEFAULT :flock);

our $VERSION = '0.2.0';


### Configuration start ###
# absolute uri path to access documentation root:
my $DOC = '/doc/';
# path to source .txt files:
my $SRC = '../.lib/doc/';       # absolute or relative to doc.cgi
# path to writable directory (it will be used for caching generated .html):
my $TMP = '../.lib/tmp/doc/';   # absolute or relative to doc.cgi
# shell command to run for converting .txt to .html (filename will be appended):
my $CMD = 'asciidoc -a toc -a icons -a linkcss -a iconsdir=/doc/images/icons -a stylesdir=/doc -a scriptsdir=/doc';
### Configuration end ###


my $doc = uri2path($ENV{REQUEST_URI});
$TMP .= $1, $SRC .= $1 if $doc =~ s{(.*/)}{};
if (grep {-f && -r && -s} "$SRC$doc.txt") {
    system("mkdir -p \Q$TMP\E");
    # two doc.cgi shouldn't process same .txt file simultaneously
    sysopen(LOCK,"$TMP$doc.lock",O_RDWR|O_CREAT)or die "open $TMP$doc.lock: $!";
    flock(LOCK, LOCK_EX)                        or die "flock: $!";
    if (grep {!-f || !-r || !-s || -M $_ > -M "$SRC$doc.txt"} "$TMP$doc.html") {
        # create/recreate symlink to source file in working directory
        unlink "$TMP$doc.txt";
        my $tmp2src = File::Spec->abs2rel($SRC, $TMP);
        symlink "$tmp2src/$doc.txt", "$TMP$doc.txt" or die "symlink: $!";
        if (system "cd \Q$TMP\E; $CMD \Q$doc.txt\E") {
            # $CMD can leave partially generated xml/html files
            unlink "$TMP$doc.$_" for qw(xml html);
            die "system: $?";
        }
    }
} else {
    # cleanup cache in case source .txt file was removed
    unlink "$TMP$doc.$_" for qw(txt xml html lock);
}
sendfile("$TMP$doc.html");


sub uri2path {
    my $s = shift;
    $s =~ s/%([0-9a-fA-F]{2})/chr(hex($1))/ge;
    return $1.'index'   if $s=~m{\A\Q$DOC\E((?:.*/)?       )\z}x;
    return $1           if $s=~m{\A\Q$DOC\E((?:.*/)? [^.]+ )\z}x;
    return $1           if $s=~m{\A\Q$DOC\E(.*) \.html?     \z}x;
    not_found();
}
sub sendfile {
    not_found() if !-f $_[0] || !-r $_[0] || !-s $_[0];
    print "Status: 200 OK\r\n";
    print "Pragma: no-cache\r\n";
    print "Expires: access\r\n";
    print "Content-Type: text/html\r\n\r\n";
    exec('cat', $_[0]);
    die "exec: $!";
}
sub not_found {
    print "Status: 404 Not Found\r\n";
    print "Content-Type: text/plain\r\n\r\n";
    print "Not found: $ENV{REQUEST_URI}\n";
    exit;
}

